static const char rcsid[] = "$Id: execComedi.c,v 1.1 2009/08/03 01:59:39 elaine Exp $";


/* $Header: /home/elaine/projeto/codigo/pronto/execComedi.c,v 1.1 2009/08/03 01:59:39 elaine Exp $
 * $Date: 2009/08/03 01:59:39 $
 *
 * Universidade do Estado do Rio de Janeiro
 * Faculdade de Engenharia 
 * Departamento de Sistemas e Computacao
 * Autora: Elaine de M. Silva
 * Projeto de Graduacao - Adequacao do Software 
 * de Simulacao Scicos para a Aquisicao de Dados
 * e o Controle em Tempo Real
 * -------------------------------------------------------------------------------------------
 * Parte do modulo servidor de funcoes de aquisicao da API do projeto COMEDI
 * 
 * $Log: execComedi.c,v $
 * Revision 1.1  2009/08/03 01:59:39  elaine
 * Initial revision
 *
 * 
 * -------------------------------------------------------------------------------------------
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <comedilib.h>
#include "comedisrv.h"


//execComedi: executa comando comedi e envia resultados
void execComedi(int newSockfd,char **par){

#define MAX_SIZE          1024   //tamanho maximo das mensagens enviadas 

#define OPEN                1   //comedi_open()
#define CONFIG              2   //comedi_config()
#define CLOSE               3   //comedi_close() 
#define GETNSUBDEV          5   //comedi_get_n_subdevices()
#define DIOREAD             6   //comedi_dio_read()
#define DIOWRITE            7   //comedi_dio_write()
#define DIOCONFIG           8   //comedi_dio_config()
#define DIOBITFIELD         9   //comedi_dio_bitfield()
#define DATAREAD           10   //comedi_data_read()
#define DATAWRITE          11   //comedi_data_write()
#define GETMAXDATA         12   //comedi_get_maxdata()
#define TOPHYS             13   //comedi_to_phys()
#define FROMPHYS           14   //comedi_from_phys()
#define DOINSN             15   //comedi_do_insn()  
#define COMMAND            16   //comedi_command()
#define COMMANDTEST        17   //comedi_command_test()
#define COMMSRCMASK        18   //comedi_get_cmd_src_mask()
#define FINDSUBDEVBYTYPE   19   //comedi_find_subdevice_by_type()
#define MODPROBE           20   //modprobe


#define MOD_PATH "/sbin/modprobe"
#define CONF_PATH "/usr/local/sbin/comedi_config"


	static comedi_t *comedi_device;
	lsampl_t data, maxdata;
	unsigned int subdevice,channel;
	int range, aref, errno;
	char *option = (char *) malloc(MAX_SIZE);
	char *driver = (char *) malloc(MAX_SIZE);
	char *model  = (char *) malloc(MAX_SIZE);



	pid_t pid;           /* ID do processo filho */
	pid_t wpid;          /* ID do processo de wait() */
	int wstatus;         /* status do wait() */

	

      	int funcao, sentBytes;
	const char *dev = "/dev/comedi";
	char arg[20];
	FILE *debug = fopen("debug.txt","w");
	if (!debug){
		fprintf(stderr, "%s: Nao foi possivel abrir %s para leitura.\n",strerror(errno),"debug.txt");
		exit(13);
	}
	
	
	//inserir tabela de relacoes
	if(!strcmp(par[1],"comedi_open")){
		funcao = OPEN;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_close")){
		funcao = 3;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_config")){
		funcao = 2;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_strerror")){
		funcao = 4;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_get_n_subdevices")){
		funcao = 5;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_dio_read")){
		funcao = 6;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_dio_write")){
		funcao = 7;
		printf("funcao = %s\n",par[1]);
	}
	else if(!strcmp(par[1],"comedi_data_read")){
		funcao = 10;
		printf("funcao = %s\n",par[1]);		
			
	}
	else if(!strcmp(par[1],"comedi_get_maxdata")){
		funcao = 12;
		printf("funcao = %s\n",par[1]);
		
	}else if(!strcmp(par[1],"comedi_modprobe")){
		funcao = 20;
		printf("funcao = %s\n",par[1]);
	}
	else{
		printf("servidor: a funcao %s nao eh suportada por este servidor ou nao existe.\n",par[1]);
		
	}
	switch(funcao){
		case OPEN:; //so pode haver 1 argumento
			  strcpy(arg,dev);
			  strcat(arg,par[2]); //concatena dispositivo e argumento para formar "/dev/comediN"
			  arg[13] = '\0';
		
			  comedi_device = comedi_open(arg);
			   if(comedi_device == NULL){
				  const char *errorMsg = (char *) malloc(sizeof(char) * MAX_SIZE);
				  if(errorMsg != NULL)
					  errorMsg = comedi_strerror(comedi_errno());
				  
				  sentBytes = sendMsg(errorMsg,newSockfd, strlen(errorMsg)); 
				  close(newSockfd);
				  }
			   else{
				  char *okMsg = "Comedi input opened successfully"; 
				  sentBytes = sendMsg(okMsg,newSockfd, strlen(okMsg));
				  if(sentBytes < 0){
					  perror("servidor:write");
					  close(newSockfd);
				  }
				  close(newSockfd);
				  }
			  break;

		case DATAREAD:;
			      int datareadstat = 0;

			      if(comedi_device != NULL){
				      subdevice = atoi((par[3]));
				      channel = atoi((par[4]));
				      range = atoi((par[5]));
				      //aref = atoi((par[6]));
				      aref = AREF_GROUND;
				      datareadstat = comedi_data_read(comedi_device,subdevice,channel,range,aref,&data);
				     
				      if (datareadstat < 0){
					     const char *dataErr = (char *) malloc(sizeof(char) * MAX_SIZE);
						if(dataErr != NULL)
					     dataErr = comedi_strerror(comedi_errno());
					     sentBytes = sendMsg(dataErr,newSockfd, strlen(dataErr));	
					     if(sentBytes < 0){
						     perror("servidor: write");
					     }
					     close(newSockfd);
						//free(dataErr);
				      }else{
					     char *dataMsg = (char *) malloc(sizeof(char) * MAX_SIZE);
					     
					     if(dataMsg != NULL)
						     snprintf(dataMsg,MAX_SIZE,"%d",data); //insere o conteudo de data na string dataMsg
					     
					     sentBytes = sendMsg(dataMsg,newSockfd,strlen(dataMsg));
					     if(sentBytes < 0){
						     perror("servidor: write");
						     }
					      close(newSockfd);

					      free(dataMsg);
					      dataMsg = NULL;
					      
					      
				      }
			      }else{
				      char *nullDevErr = "servidor: o dispositivo comedi nao esta aberto";
				      sentBytes = sendMsg(nullDevErr,newSockfd,strlen(nullDevErr));
                                      close(newSockfd);
			      }

			  break;

		case GETMAXDATA:;
				
				if(comedi_device != NULL){
					subdevice = atoi((par[3]));
					channel = atoi((par[4]));
					maxdata = comedi_get_maxdata(comedi_device, subdevice, channel);
					if(maxdata == 0){
						const char * maxErr = (char *) malloc(sizeof(char) * MAX_SIZE);
						if (maxErr != NULL)
							maxErr = comedi_strerror(comedi_errno());
						
						sentBytes = sendMsg(maxErr,newSockfd,strlen(maxErr));
						if(sentBytes < 0){
							perror("servidor: write");
						}
						printf("servidor: erro em comedi_get_maxdata\n");	
										 
						close(newSockfd);
					}else{

						char *Maxmsg = (char *) malloc(sizeof(char) * MAX_SIZE);
						if (Maxmsg != NULL)
							snprintf(Maxmsg,MAX_SIZE,"%d",maxdata);

						sentBytes = sendMsg(Maxmsg,newSockfd,strlen(Maxmsg));
						if(sentBytes < 0){
							perror("servidor: write");
						}
						
						free(Maxmsg);
						Maxmsg = NULL;
						
						close(newSockfd);
					}
				}else{
					printf("servidor: O dispositivo comedi nao esta aberto\n");
					close(newSockfd);
				}
			  break;

		case GETNSUBDEV:;
				if(comedi_device != NULL){
					int nsubdev;
					nsubdev = comedi_get_n_subdevices(comedi_device);
					if (nsubdev < 0){
						const char *subdevErr =(char *)  malloc(sizeof(char) * MAX_SIZE);
						if (subdevErr != NULL)
							subdevErr = comedi_strerror(comedi_errno());
						
						sentBytes = sendMsg(subdevErr,newSockfd,strlen(subdevErr));
						if (sentBytes < 0){
							perror("servidor: write");
						}
						
						close(newSockfd);
						}else{
							char *nsubdevMsg =(char *)  malloc(sizeof(char) * MAX_SIZE);
							if (nsubdevMsg != NULL)
								snprintf(nsubdevMsg,MAX_SIZE,"%d",nsubdev);

							sentBytes = sendMsg(nsubdevMsg,newSockfd,strlen(nsubdevMsg));
							if(sentBytes < 0){
								perror("servidor: write");
							}
								
							free(nsubdevMsg);
							nsubdevMsg = NULL;

							
							close(newSockfd);
						}
				}else{
						printf("servidor: O dispositivo comedi nao esta aberto\n");
						close(newSockfd);
						}
					break;




		case MODPROBE:;

			     int uidStatusSU, uidStatusU;
			      
			      static int execStatusMod;
			      
				if(option != NULL)
				      strcpy(option, "driver");
			      
			      if (strncmp(par[2], "-",1) != 0){ /* testa opcao do comando modprobe */ 
				        printf("servidor: Digite a opcao do comando modprobe (-v, -r, -c, -n...consulte o manual do comando para mais informacoes)\n");
				      break;
			      
			      }
if(driver != NULL)
			      getDriverName(par[3],MAX_SIZE, option, driver); /* pega nome do driver no arquivo de configuracao */
			      			       
			      if (driver == NULL){ /* testa se foi encontrado driver a partir do modelo no arq. de conf */
				      printf("servidor: Nao foi possivel encontrar um driver para o dispositivo %s\n",par[3]);
				      break;
			      }

			      
			      static char *argsMod[4] ;
			      argsMod[0] = "modprobe";
			      argsMod[1] = par[2];
			      argsMod[2] = driver;
			      argsMod[3] = NULL;


			      /*
			       *
			       * Inicia bifurcacao em dois processos para chamar o modprobe na linha de comando
			       * do processo filho. Primeiro  chamado o comando fork() para criar um processo
			       * filho. Dentro do filho  chamado o comando setuid(0) para dar privilegios
			       * de superusuario necessarios a execucao do comando modprobe. Logo aps sao chamados
			       * os comandos close(n). Eles fecham as saidas padrao e erro padrao e o comando dup
			       * redireciona essas saidas para o socket. Com isso as mensagens de erro e de saida
			       * sao redirecionadas para o cliente. O modprobe  executado pela funcao execCmd e 
			       * o UID voltara para o status de usuario comum. Enquanto isso o processo pai espera 
			       * o retorno do processo filho. Por fim, este retorno  testado pelo status de wait().
			       *
			       */


			      pid = fork();

			      if (pid == -1){
				      /*
				       * Erro no fork():
				       */
				      fprintf(stderr,"%s: Failed to fork()\n", strerror(errno));
				      exit(13);
			      } else if (pid == 0){
				      /* 
				       * Processo filho executando:
				       */
				      printf("PID %ld: Processo filho iniciado,processo pai eh: %ld. \n",
						      (long) getpid(),   /* processo filho */
						      (long) getppid()); /* processo pai   */

				      uid_t uIDsu = 0;        /* seta ID de superusuario */
				      uid_t uIDu = getuid();  /* pega ID de usuario */ 
				      uidStatusSU = setuid(uIDsu); /* seta uid de superusuario para executar modprobe */
				     
				      if(uidStatusSU < 0){   /* confere execucao do setuid para superusuario */
					      fprintf(debug, "%s: Failed to set uid\n",strerror(errno));
					      fflush(debug);
					      close(newSockfd);
					      }
				      else{
					      close(1); /* fecha erro e saida padrao e redireciona para socket */
					      close(2);  
					      dup2(newSockfd,1);
					      dup2(newSockfd,2);

					      fprintf(debug,"parametros do modprobe: '%s' e '%s'\n",par[2], driver);
					      fflush(debug);

					      execStatusMod =  execCmd(MOD_PATH,argsMod);

					      if(execStatusMod == 1){

						      printf("servidor: Erro no nome do driver\n"); 
						      close(newSockfd);
						      }

					      uidStatusU = setuid(uIDu); /* volta com uid de usuario */
					      if(uidStatusU < 0){
						      fprintf(debug,"%s: Failed to set uid\n",strerror(errno)); 
						      fflush(debug);
						      close(newSockfd);
						      }
					      }
				     
					      }
			      else {
				      /* 
				       * Processo pai executando:
				       */
				      printf("PID %ld: Processo filho inicia no PID %ld. \n",
						      (long)getpid(), /* processo pai */
						      (long) pid);    /* PID do filho */

				      /*
				       * Espera fim do processo filho:
				       */
				      wpid = wait(&wstatus);  /* Status de saida do processo filho */
				      if ( wpid == -1 ){
					      /* 
					       * Se a chamada a wait() falhar: 
					       */
					      fprintf(stderr,"%s: wait() \n", strerror(errno));
					     // return 1;
				      }
				      else if (wpid != pid)
					      /* Esta situacao nunca deve ocorrer! */
					      abort();
				      else{
					      /*
					       * Processo filho terminou:
					       */
					      if (WIFEXITED(wstatus)){ /* Usando as macros para teste do wait */
						     /*
						      * Saida normal
						      */
						     printf("Saiu: $? = %d\n", WEXITSTATUS(wstatus));
					      } else if( WIFSIGNALED(wstatus) ){
						     /*
						      * Se o processo abortou, foi sinalizado ou eliminado:
						      */
						     printf("Sinal : %d%s\n",WTERMSIG(wstatus), WCOREDUMP(wstatus) ? " com arquivo de core. " : "");
					      }else {
						     /* 
						      * Processo filho interrompido:
						      */
						     printf("Parado.\n");
					      }
				      }
			      }


			      free(option);
			      free(driver);
			      free(model);

			      option = NULL;
			      driver = NULL;
			      model = NULL;

			      break;


		case CONFIG:;
			if(option != NULL)
			    strcpy(option, "model");

			    bzero(arg,sizeof arg);				
			    strcpy(arg,dev);
			    strcat(arg,par[3]);


			    if (strncmp(par[2], "-",1) != 0){ /* testa opcao do comando comedi_config */ 
				    printf("servidor: Digite a opcao do comando modprobe (-v, -q, -i, -r...digite 'comedi_config --help'  para mais informacoes)\n");
				    break;

				    }

			    printf("Argumentos: %s,%s,%s\n",arg,par[3],par[4]);
			    if(model != NULL)
				    getDriverName(par[4],MAX_SIZE,option, model);

			    printf("Modelo: %s\n", model);

			    static char *argsConf[5] ;
			    argsConf[0] = "comedi_config";
			    argsConf[1] = par[2];
			    argsConf[1] = arg;
			    argsConf[2] = model;
			    argsConf[3] = par[5];
			    argsConf[4] = NULL;

			    /*
			     *
			     * Inicia bifurcacao em dois processos para chamar o modprobe na linha de comando
			     * do processo filho. Primeiro  chamado o comando fork() para criar um processo
			     * filho. Dentro do filho  chamado o comando setuid(0) para dar privilegios
			     * de superusuario necessarios a execucao do comando modprobe. Logo aps sao chamados
			     * os comandos close(n). Eles fecham as saidas padrao e erro padrao e o comando dup
			     * redireciona essas saidas para o socket. Com isso as mensagens de erro e de saida
			     * sao redirecionadas para o cliente. O modprobe  executado pela funcao execCmd e 
			     * o UID voltara para o status de usuario comum. Enquanto isso o processo pai espera 
			     * o retorno do processo filho. Por fim, este retorno  testado pelo status de wait().
			     *
			     */

			    pid = fork();
			    if (pid == -1){
				    
				    /*
				     * Erro no fork():
				     */

				    fprintf(stderr,"%s: Failed to fork()\n", strerror(errno));
				    exit(13);

				    } else if (pid == 0){
				      /* 
				       * Processo filho executando:
				       */
				      printf("PID %ld: Processo filho iniciado,processo pai eh: %ld. \n",
						      (long) getpid(),   /* processo filho */
						      (long) getppid()); /* processo pai   */
			  
				    close(1); //fecha erro e saida padrao e redireciona para socket
				    close(2);
				    dup2(newSockfd,1);
				    dup2(newSockfd,2);

				    //executa comando comedi_config. O argumento obrigatorio eh o dispositivo e modelo
				    //Placas pnp nao precisam argumentos adicionais. Placas nao pnp precisam de endereco 
				    //de IO base e IRQ. Argumentos opcionais sao canal DMA, velocidade do clock principal,
				    //faixa de entrada analogica (menor e maior tensao, em microvolts), faixa de saida 
				    //analogica (menor e maior tensao, em microvolts), modo do timer para o DMA.
				    //ex.: comedi_config /dev/comedi10 comedi_parport 0x370,5 (IO Base,IRQ)
				    //     comedi_config /dev/comedi1 das16 0x390,3,3,10 (IO Base, IRQ,DMA Mode, Clock).  
				    //Como o tokenizer nao separa por virgulas, a string de opcionais eh toda tratada por par[4].
				    

				    execStatusMod =  execCmd(CONF_PATH,argsConf);
				   
				    }else {
				      /* 
				       * Processo pai executando:
				       */
				      printf("PID %ld: Processo filho inicia no PID %ld. \n",
						      (long)getpid(), /* processo pai */
						      (long) pid);    /* PID do filho */

				      /*
				       * Espera fim do processo filho:
				       */
				      wpid = wait(&wstatus);  /* Status de saida do processo filho */
				      if ( wpid == -1 ){
					      /* 
					       * Se a chamada a wait() falhar: 
					       */

					      fprintf(stderr,"%s: wait() \n", strerror(errno));
					      
					      }

				      else if (wpid != pid)

					      /* Esta situacao nunca deve ocorrer! */
					      abort();
				      else{
					      /*
					       * Processo filho terminou:
					       */
					      if (WIFEXITED(wstatus)){ /* Usando as macros para teste do wait */
						     /*
						      * Saida normal
						      */
						     printf("Saiu: $? = %d\n", WEXITSTATUS(wstatus));
					      } else if( WIFSIGNALED(wstatus) ){
						     /*
						      * Se o processo abortou, foi sinalizado ou eliminado:
						      */
						     printf("Sinal : %d%s\n",WTERMSIG(wstatus), WCOREDUMP(wstatus) ? " com arquivo de core. " : "");
					      }else {
						     /* 
						      * Processo filho interrompido:
						      */
						     printf("Parado.\n");
					      }
				      }
			      }


			    free(option);
			    free(driver);
			    free(model);

			    option = NULL;
			    driver = NULL;
			    model = NULL;

			    fclose(debug);

			    break;

			    }

}
